from cspy.algorithms.bidirectional import BiDirectional
from cspy.algorithms.tabu import Tabu
from cspy.preprocessing import check_and_preprocess
from cspy.label import Label

name = "cspy"
__all__ = ['BiDirectional', 'Tabu', 'check_and_preprocess', 'Label']



"""TODO:
 -  write tests for Tabu algorithm
"""

import sys
import unittest
import networkx as nx
sys.path.append("../")
from cspy.algorithms.bidirectional import BiDirectional
from cspy.algorithms.tabu import Tabu
from cspy.label import Label


class TestsBasic(unittest.TestCase):
    """ Tests for finding the resource constrained shortest
    path of simple DiGraph using the BiDirectional algorithm."""

    def setUp(self):
        self.max_res, self.min_res = [4, 20], [1, 0]
        # Create simple digraph to test algorithm
        self.G = nx.DiGraph(directed=True, n_res=2)
        self.G.add_edge('Source', 'A', res_cost=[1, 2], weight=0)
        self.G.add_edge('A', 'B', res_cost=[1, 0.3], weight=0)
        self.G.add_edge('A', 'C', res_cost=[1, 0.1], weight=0)
        self.G.add_edge('B', 'C', res_cost=[1, 3], weight=-10)
        self.G.add_edge('B', 'Sink', res_cost=[1, 2], weight=10)
        self.G.add_edge('C', 'Sink', res_cost=[1, 10], weight=0)

        # Create erratic digraph to test exception handling
        self.E = nx.DiGraph(directed=True)
        self.E.add_edge('Source', 'A', weight=0)

        # Create digraph with negative resource costs with unreachable node 'B'
        self.H = nx.DiGraph(directed=True, n_res=2)
        self.H.add_edge('Source', 'A', res_cost=[1, 2], weight=0)
        self.H.add_edge('A', 'C', res_cost=[-1, 0.3], weight=0)
        self.H.add_edge('A', 'B', res_cost=[-1, 3], weight=0)
        self.H.add_edge('B', 'D', res_cost=[-1, 2], weight=0)
        self.H.add_edge('C', 'D', res_cost=[1, 0.1], weight=0)
        self.H.add_edge('D', 'Sink', res_cost=[1, 0.1], weight=0)

        self.J = nx.DiGraph(directed=True, n_res=2)
        self.J.add_edge('Source', 'A', res_cost=[1, 1], weight=1)
        self.J.add_edge('Source', 'B', res_cost=[1, 1], weight=1)
        self.J.add_edge('A', 'C', res_cost=[1, 1], weight=1)
        self.J.add_edge('B', 'C', res_cost=[2, 1], weight=-1)
        self.J.add_edge('C', 'D', res_cost=[1, 1], weight=-1)
        self.J.add_edge('D', 'E', res_cost=[1, 1], weight=1)
        self.J.add_edge('D', 'F', res_cost=[1, 1], weight=1)
        self.J.add_edge('F', 'Sink', res_cost=[1, 1], weight=1)
        self.J.add_edge('E', 'Sink', res_cost=[1, 1], weight=1)

    def testBothDirections(self):
        # Find shortest path of simple test digraph
        path = BiDirectional(self.G, self.max_res, self.min_res).run()
        self.assertEqual(path, ['Source', 'A', 'B', 'C', 'Sink'])

    def testForward(self):
        # Find shortest path of simple test digraph
        path = BiDirectional(
            self.G, [200, 20], self.min_res, 'forward').run()
        self.assertEqual(path, ['Source', 'A', 'B', 'C', 'Sink'])

    def testBackward(self):
        # Find shortest path of simple test digraph
        path = BiDirectional(self.G, self.max_res, [-1, 0], 'backward').run()
        self.assertEqual(path, ['Source', 'A', 'B', 'C', 'Sink'])

    def testDominance(self):
        # Check whether forward and backward dominance makes sense
        L1 = Label(10, 'B', [6, 5], [])
        L2 = Label(1, 'B', [6, -3], [])
        L3 = Label(-10, 'A', [3, -8.3], [])
        L4 = Label(-9, 'A', [4, -6.3], [])
        L5 = Label(0, 'A', [4, -5.1], [])
        self.assertTrue(L2.dominates(L1))
        self.assertTrue(L3.dominates(L4))
        self.assertTrue(L3.dominates(L5))

    def testExceptions(self):
        # Check whether wrong input raises exceptions
        self.assertRaises(Exception, BiDirectional, self.E,
                          'x', [1], 'up')

    def testNegativeEdges(self):
        # Check if negative resource costs work and whether
        # unreachable nodes are eliminated
        path = BiDirectional(self.H, [5, 20], [0, 0]).run()
        # check if the unreachable node has been eliminated
        self.assertTrue('B' not in self.H.nodes())
        self.assertEqual(path, ['Source', 'A', 'C', 'D', 'Sink'])

    def testTabu(self):
        path = Tabu(self.J, [5, 5], [0, 0]).run()
        self.assertEqual(path, ['Source', 'A', 'C', 'D', 'E', 'Sink'])


if __name__ == '__main__':
    unittest.main()



import logging
import networkx as nx


def check(G, max_res=None, min_res=None, direc_in=None):
    """Check whether inputs have the appropriate attributes and
    are of the appropriate types."""

    def _check_res():
        if isinstance(max_res, list) and isinstance(min_res, list):
            if len(max_res) == len(min_res) >= 2:
                if (all(isinstance(i, (float, int)) for i in max_res) and
                        all(isinstance(i, (float, int)) for i in min_res)):
                    pass
                else:
                    raise Exception("Elements of input lists must be numbers")
            else:
                raise Exception("Input lists have to be equal length >= 2")
        else:
            raise Exception("Inputs have to be lists with length >= 2")

    def _check_direction():
        if direc_in not in ['forward', 'backward', 'both']:
            raise Exception(
                "Input direction has to be 'forward', 'backward', or 'both'")

    def _check_graph_attr():
        """Checks whether input graph has n_res attribute"""
        if isinstance(G, nx.DiGraph):
            if 'n_res' not in G.graph:
                raise Exception("Input graph must have 'n_res' attribute.")
        else:
            raise Exception("Input must be a nx.Digraph()")

    def _check_edge_attr():
        """Checks whether edges in input graph have res_cost attribute"""
        if not all('res_cost' in edge[2] for edge in G.edges(data=True)):
            raise Exception(
                "Input graph must have edges with 'res_cost' attribute.")

    def _check_path():
        """Checks whether a 'Source' -> 'Sink' path exists.
        Also covers nodes missing and other standard networkx exceptions"""
        try:
            nx.has_path(G, 'Source', 'Sink')
        except nx.NetworkXException as e:
            raise Exception("An error occured: {}".format(e))

    errors = []
    # Select checks to perform based on the input provided
    if max_res and min_res and direc_in:
        check_funcs = [_check_res, _check_direction, _check_graph_attr,
                       _check_edge_attr, _check_path]
    elif max_res and min_res:
        check_funcs = [_check_res, _check_graph_attr,
                       _check_edge_attr, _check_path]
    else:
        check_funcs = [_check_path]
    # Check all functions in check_funcs
    for func in check_funcs:
        try:
            func()
        except Exception as e:
            errors.append(e)  # if check fails save error message
    if errors:
        # if any check has failed raise an exception with all the error
        # messages
        raise Exception('\n'.join('{}'.format(item) for item in errors))


def prune_graph(G, max_res, min_res):
    """Removes nodes that cannot be reached due to resource limits.

    Parameters
    ----------

    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

     max_res : list of floats
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats
        :math:`[U, L_1, L_2, ..., L_{n\_res}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    Raises
    ------

    Raises exceptions if incorrect input is given. If multiple exceptions are
    raised, and exception with a list of exceptions is raised.

    Returns
    -------
    Preprocessed graph G
    """

    def _check_resource(r):
        # check resource r's feasibility along a path

        def _get_weight(i, j, attr_dict):
            # returns number to use as weight for the algorithm
            return attr_dict['res_cost'][r]

        # Get paths from source to all other nodes
        length, path = nx.single_source_bellman_ford(
            G, 'Source', weight=_get_weight)
        try:
            # If any path violates the resource upper or lower bounds
            # then, add the problematic node to the dictionary
            nodes_to_remove.update({
                path[key][-2]: val for key, val in length.items()
                if val > max_res[r] or val < min_res[r]})
            if "Source" in nodes_to_remove or "Sink" in nodes_to_remove:
                raise Exception("Sink not reachable for resource {}".format(r))
            # path is a dict of the form:
            # {node_i: [Source, ..., node_k, node_i]}.
            # Hence, if node_i is found to violate a resource constraint, it is
            # because of node_k, therefore, we add path[key][-2] = node_k to
            # the dictionary of nodes to remove.
        except IndexError:  # No nodes violate resource limits
            pass

    nodes_to_remove = {}
    # Map function for each resource
    list(map(_check_resource, range(0, G.graph['n_res'])))
    if nodes_to_remove:  # if there are nodes to remove
        # Filter out source or sink
        nodes_to_remove = {
            key: val for key, val in nodes_to_remove.items()
            if key != 'Source' or key != 'Sink'}
        G.remove_nodes_from(nodes_to_remove)
        logging.info("[{0}] Removed {1} nodes".format(
            __name__, len(nodes_to_remove)))
    return G


def check_and_preprocess(preprocess, G, max_res=None, min_res=None,
                         direction=None):
    """
    Checks whether inputs and the graph are of the appropriate types and
    have the required properties.
    Removes nodes that cannot be reached due to resource limits.

    Parameters
    ----------
    ``preprocess`` : bool
        enables preprocessing routine.

    ``G`` : object instance ``nx.Digraph()``
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    ``max_res`` : list of floats, optional
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    ``min_res`` : list of floats, optional
        :math:`[U, L_1, L_2, ..., L_{nres}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    ``direction`` : string, optional
        preferred search direction. Either 'both','forward', or, 'backward'.
        Default : 'both'.

    :return: If ``preprocess``, returns preprocessed graph ``G`` if no
        exceptions are raised, otherwise doesn't return anything.

    :raises: Raises exceptions if incorrect input is given. If multiple exceptions are
        raised, and exception with a list of exceptions is raised.
    """
    check(G, max_res, min_res, direction)
    if preprocess:
        G = prune_graph(G, max_res, min_res)
        check(G)
        return G



from operator import sub


class Label(object):
    """Label object that allows comparison and the modelling of dominance relations

    Parameters
    -----------

    weight : float
        cumulative edge weight

    node : string
        name of last node visited

    res : list
        cumulative edge resource consumption

    path : list
        all nodes in the path"""

    _REF_forward, _REF_backward = None, None

    def __init__(self, weight, node, res, path):
        self.weight = weight
        self.node = node
        self.res = res
        self.path = path

    def __repr__(self):
        return str(self)

    def __str__(self):  # for printing purposes
        return "Label({0},{1},{2})".format(self.weight, self.node, self.res)

    def __lt__(self, other):
        # Less than operator for two Label objects
        return (self.weight < other.weight and self.res <= other.res) or (
            self.weight <= other.weight and self.res < other.res)

    def __le__(self, other):
        # Less than or equal to operator for two Label objects
        return self.weight <= other.weight and self.res <= other.res

    def __eq__(self, other):
        # Equality operator for two Label objects
        return (self.weight == other.weight and self.res == other.res and
                self.node == other.node)

    def __hash__(self):
        # Redefinition of hash to avoid TypeError due to the __eq__ definition
        return id(self)

    def dominates(self, other, direction="forward"):
        # Return whether self dominates other.
        if direction == 'forward':
            return self.node == other.node and (
                (self.weight < other.weight and self.res <= other.res) or (
                    self.weight <= other.weight and self.res < other.res))
        else:
            return self.node == other.node and (
                (self.weight < other.weight and self.res >= other.res) or (
                    self.weight <= other.weight and self.res > other.res))

    def get_new_label(self, direction, weight, node, res):
        path = list(self.path)
        path.append(node)
        if direction == 'forward':
            res_new = list(map(self._REF_forward, self.res, res))
        else:
            res_new = list(map(self._REF_backward, self.res, res))
        return Label(weight + self.weight, node, res_new, path)

    def feasibility_check(self, max_res=[], min_res=[],
                          direction="forward"):
        if direction == "forward":
            return self.check_geq(max_res, self.res)
        else:
            return self.check_geq(self.res, min_res, "gt")

    @staticmethod
    def check_geq(l1, l2, inequality="ge"):
        """Determines if all elements of list l1 either >=
        or > than those in list l2.

        Parameters
        ----------
        l1, l2 : list
            lists of integers.

        inequality : string, optional
            type of inequality 'ge' for >= 'gt' for >. Default: 'ge'
            """
        diff = list(map(sub, l1, l2))
        if inequality == "gt":
            return all(elem > 0 for elem in diff)
        else:
            return all(elem >= 0 for elem in diff)



from __future__ import absolute_import
from __future__ import print_function

import logging
import numpy as np
import networkx as nx
from cspy.label import Label
from cspy.preprocessing import check


class Tabu:
    """Heuristic Tabu search algorithm for the (resource) constrained shortest
    path problem.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

     max_res : list of floats
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats
        :math:`[U, L_1, L_2, ..., L_{n\_res}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    Returns
    -------
    path : list
        nodes in shortest path obtained.

    Notes
    -----
    The input graph must have a ``n_res`` attribute in the input graph has
    to be :math:`\geq 2`. The edges in the graph must all have a `res_cost`
    attribute.

    Example
    -------
    To run the algorithm, create a :class:`Tabu` instance and call `run`.

    .. code-block:: python

        >>> from cspy import Tabu
        >>> G = nx.DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=[1, 1], weight=1)
        >>> G.add_edge('Source', 'B', res_cost=[1, 1], weight=1)
        >>> G.add_edge('A', 'C', res_cost=[1, 1], weight=1)
        >>> G.add_edge('B', 'C', res_cost=[2, 1], weight=-1)
        >>> G.add_edge('C', 'D', res_cost=[1, 1], weight=-1)
        >>> G.add_edge('D', 'E', res_cost=[1, 1], weight=1)
        >>> G.add_edge('D', 'F', res_cost=[1, 1], weight=1)
        >>> G.add_edge('F', 'Sink', res_cost=[1, 1], weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=[1, 1], weight=1)
        >>> max_res, min_res = [5, 5], [0, 0]
        >>> path = Tabu(G, max_res, min_res).run()
        >>> print(path)
        ['Source', 'A', 'B', 'C', 'Sink']

    """

    def __init__(self, G, max_res, min_res):
        # Input parameters
        check(G, max_res, min_res)
        self.G = G
        self.max_res = max_res
        self.min_res = min_res
        # Algorithm specific parameters
        self.it = 0
        self.path = []
        self.stop = False
        self.H = self.G.copy()  # subgraph
        self.predecessor_edges = []
        self.edges_to_remove = dict(self.G.edges())
        # Import function from cspy.label.Label
        self.check_feasibility = Label.check_geq

    def run(self):
        """Runs Tabu Search with resource constraints for the input graph"""
        while self.stop is False:
            self.algorithm()
            self.it += 1

        if self.path:
            logging.debug(self.path)
            return self.path
        else:
            raise Exception("No resource feasible path has been found")

    def algorithm(self):
        path = []
        try:
            path = nx.astar_path(self.H, 'Source', 'Sink')
        except nx.NetworkXException:
            pass
        self.H = self.G.copy()  # subgraph
        if path:
            shortest_path_edges = (edge for edge in
                                   self.G.edges(
                                       self.G.nbunch_iter(path), data=True)
                                   if edge[0:2] in zip(path, path[1:]))
            total_res = np.zeros(self.G.graph['n_res'])
            for edge in shortest_path_edges:
                total_res += self._edge_extract(edge)
                if (self.check_feasibility(self.max_res, total_res) and
                        self.check_feasibility(total_res, self.min_res)):
                    pass
                else:
                    self._update_graph(edge)
                    break
            else:
                self.path = path
                self.stop = True
        else:
            self._update_graph()

    def _update_graph(self, edge=None):
        if self.it == 0:
            self.last_edge_removed = edge
        if self.predecessor_edges:
            edge_to_remove = self.next_predecessor_edge()
        else:
            if self.edges_to_remove:
                self.get_predecessor_edges(self.last_edge_removed)
                if edge and edge[:2] in self.edges_to_remove:
                    edge_to_remove = edge
                elif self.predecessor_edges:
                    edge_to_remove = self.next_predecessor_edge()
                else:
                    self.stop = True
                    return
            else:
                self.stop = True
                return
        self.edges_to_remove.pop(edge_to_remove[:2], None)
        self.H.remove_edge(*edge_to_remove[:2])
        self.last_edge_removed = edge_to_remove

    def get_predecessor_edges(self, edge):
        node = edge[0]
        self.predecessor_edges = [
            e for e in self.G.edges(
                self.G.nbunch_iter([node] + list(self.G.predecessors(node))),
                data=True)
            if e[1] == node and e != edge]
        self.predecessor_edges.sort(
            key=lambda x: x[2]['weight'])

    def next_predecessor_edge(self):
        """Retrieves the edge adjacent to node with the largest weight"""
        next_edge = self.predecessor_edges[-1]
        self.predecessor_edges.pop(-1)
        return next_edge

    @staticmethod
    def _edge_extract(edge):
        return np.array(edge[2]['res_cost'])



from __future__ import absolute_import
from __future__ import print_function

import random
import logging
from operator import add, sub
from collections import OrderedDict

from cspy.label import Label
from cspy.preprocessing import check_and_preprocess


class expand:
    pass


class BiDirectional:
    """
    Implementation of the bidirectional labeling algorithm with dynamic
    half-way point Tilk2017.
    Depending on the range of values for self.HF = HF and self.HB = HB, we get
    four different algorithms. See self.name_algorithm.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats
        :math:`[U, L_1, L_2, ..., L_{n\_res}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    direc_in : string, optional
        preferred search direction.
        Either 'both','forward', or, 'backward'. Default : 'both'.

    preprocess : bool, optional
        enables preprocessing routine.

    REF_forward REF_backward : functions, optional
        non-additive resource extension functions.

    Returns
    -------
    list
        nodes in shortest path obtained.

    Notes
    -----
    The input graph must have a ``n_res`` attribute in the input graph has
    to be :math:`\geq 2`. The edges in the graph must all have a ``res_cost``
    attribute.

    According to the inputs, four different algorithms can be implemented:

    - HF = HB > U or ``direc_in`` = 'forward': Monodirectional forward labeling algorithm
    - L < HF = HB < U: Bidirectional labeling algorithm with static halfway point
    - HF = HB < L or ``direc_in`` == 'backward': Monodirectional backward labeling algorithm
    - U = HF > HB = L: Bidirectional labeling algorithm with dynamic halfway point.

    Example
    -------
    To run the algorithm, create a :class:`BiDirectional` instance and call
    ``run``.

    .. code-block:: python

        >>> import cspy
        >>> import networkx as nx
        >>> G = nx.DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=[1, 2], weight=0)
        >>> G.add_edge('A', 'B', res_cost=[1, 0.3], weight=0)
        >>> G.add_edge('A', 'C', res_cost=[1, 0.1], weight=0)
        >>> G.add_edge('B', 'C', res_cost=[1, 3], weight=-10)
        >>> G.add_edge('B', 'Sink', res_cost=[1, 2], weight=10)
        >>> G.add_edge('C', 'Sink', res_cost=[1, 10], weight=0)
        >>> max_res, min_res = [4, 20], [1, 0]
        >>> algObj = BiDirectional(G, max_res, min_res, direction='both')
        >>> path = algObj.run()
        >>> print(path)
        ['Source', 'A', 'B', 'C', 'Sink']

    """

    def __init__(self, G, max_res, min_res, direction='both',
                 preprocess=True, REF_forward=add, REF_backward=sub,):

        self.G = check_and_preprocess(
            preprocess, G, max_res, min_res, direction)
        self.direc_in = direction
        self.max_res, self.min_res = max_res, min_res
        self.L, self.U = self.max_res[0], self.min_res[0]
        self.HB = self.L  # type: float
        self.HF = self.U  # type: float

        Label._REF_forward = REF_forward
        Label._REF_backward = REF_backward

        self.name_algorithm()

        n_res = G.graph['n_res']

        self.Label = {
            # init current forward label
            'forward': Label(0, 'Source', [0] * n_res, ['Source']),
            # init current backward label
            'backward': Label(0, 'Sink', max_res, ['Sink'])}

        # init forward and backward unprocessed labels
        self.unprocessed = dict()
        self.unprocessed['forward'], self.unprocessed['backward'] = {}, {}
        # init final path
        self.finalpath = dict()
        self.finalpath['forward'], self.finalpath['backward'] = [
            "Source"], ["Sink"]

    def run(self):
        while self.Label['forward'] or self.Label['backward']:
            direc = self.get_direction()
            if direc:
                self.algorithm(direc)
                self.check_dominance(direc)
            elif not direc or self.terminate(direc):
                break
        return self.join_paths()

    #############
    # DIRECTION #
    #############
    def get_direction(self):
        if self.direc_in == 'both':
            if self.Label['forward'] and not self.Label['backward']:
                return 'forward'
            elif not self.Label['forward'] and self.Label['backward']:
                return 'backward'
            elif self.Label['forward'] and self.Label['backward']:
                return random.choice(['forward', 'backward'])
            else:  # if both are empty
                return
        else:
            if not self.Label['forward'] and not self.Label['backward']:
                return
            elif not self.Label[self.direc_in]:
                return
            else:
                return self.direc_in

    #############
    # ALGORITHM #
    #############
    def algorithm(self, direc):

        def _propagate_label(edge):
            # Label propagation #
            weight, res_cost = edge[2]['weight'], edge[2]['res_cost']
            node = edge[1] if direc == 'forward' else edge[0]
            new_label = self.Label[direc].get_new_label(
                direc, weight, node, res_cost)
            if new_label.feasibility_check(
                    self.max_res, self.min_res, direc):
                self.unprocessed[direc][self.Label[direc]][
                    new_label] = new_label.path

        def _get_next_label():
            # Label Extension #
            keys_to_pop = []
            for key, val in self.unprocessed[direc].items():
                if val:
                    # Update next forward label with one with least weight
                    next_label = min(val.keys(), key=lambda x: x.weight)
                    # Remove it from the unprocessed labels
                    self.unprocessed[direc][key].pop(next_label)
                    self.Label[direc] = next_label
                    if not self.unprocessed[direc][key]:
                        keys_to_pop.append(key)
                    break
                else:
                    keys_to_pop.extend([self.Label[direc], key])
            else:  # if no break
                next_label = min(
                    self.unprocessed[direc].keys(), key=lambda x: x.weight)
                # Remove it from the unprocessed labels
                keys_to_pop.append(next_label)
                if self.Label[direc] == next_label:
                    self.save_final_path(direc)
                    keys_to_pop.append(self.Label[direc])
                    self.Label[direc] = None
                else:
                    self.Label[direc] = next_label
            # Remove all processed labels from unprocessed dict
            for k in list(set(keys_to_pop)):
                self.unprocessed[direc].pop(k, None)

        if direc == 'forward':  # forward
            idx = 0
            self.min_res[0] = max(self.min_res[0], min(
                self.Label[direc].res[0], self.max_res[0]))
        else:
            idx = 1
            self.max_res[0] = min(self.max_res[0], max(
                self.Label[direc].res[0], self.min_res[0]))
        edges = [e for e in self.G.edges(data=True)
                 if e[idx] == self.Label[direc].node]
        if self.Label[direc] not in self.unprocessed[direc]:
            self.unprocessed[direc][self.Label[direc]] = {}

        list(map(_propagate_label, edges))
        _get_next_label()

    ###############
    # TERMINATION #
    ###############
    def terminate(self, direc):
        if self.direc_in == "both":
            if (self.finalpath['forward'] and self.finalpath['backward'] and
                    (self.finalpath['forward'][-1] ==
                        self.finalpath['backward'][-1])):
                return True
        else:
            if (self.finalpath['forward'][-1] == "Sink" and
                    not self.unprocessed['forward']):
                return True
            elif (self.finalpath['backward'][-1] == "Source" and
                  not self.unprocessed['backward']):
                return True
            if not self.Label[self.direc_in] and not self.G.edges():
                return True

    def save_final_path(self, direc):
        if self.Label[direc]:
            self.finalpath[direc] = self.Label[direc].path

    #############
    # DOMINANCE #
    #############
    def check_dominance(self, direc):
        """ For all labels, check if it is dominated, or itself dominates other
        labels. If this is found to be the case, the dominated label is
        removed """
        for sub_dict in ({k: v} for k, v in self.unprocessed[direc].items()):
            k = list(sub_dict.keys())[0]  # call dict_keys object as a list
            for label in [key for v in sub_dict.values() for key in v.keys()]:
                if label.node == self.Label[direc].node:
                    if self.Label[direc].dominates(label, direc):
                        self.unprocessed[direc][k].pop(
                            label, None)
                        self.unprocessed[direc].pop(label, None)
                    elif label.dominates(self.Label[direc], direc):
                        self.unprocessed[direc][k].pop(self.Label[direc], None)
                        self.unprocessed[direc].pop(self.Label[direc], None)

    #################
    # PATH CHECKING #
    #################
    def join_paths(self):
        # check if paths are eligible to be joined

        def _check_paths():
            if (self.finalpath['forward'][-1] == 'Sink' and
                    self.finalpath['backward'][0] != 'Source'):
                # if only backward path
                return self.finalpath['forward']
            elif (self.finalpath['backward'][0] == 'Source' and
                  self.finalpath['forward'][-1] != 'Sink'):
                # if only backward path
                return self.finalpath['backward']
            elif (self.finalpath['backward'][0] == 'Source' and
                  self.finalpath['forward'][-1] == 'Sink'):
                # if both full paths
                return random.choice(
                    [self.finalpath['forward'], self.finalpath['backward']])
            elif not self.Label['forward'] or not self.Label['backward']:
                # if combination of the two is required
                return list(OrderedDict.fromkeys(
                    self.finalpath['forward'] + self.finalpath['backward']))
            else:
                return

        if self.direc_in == 'both':
            if self.finalpath['forward'] and self.finalpath['backward']:
                # reverse order for backward path
                self.finalpath['backward'].reverse()
                return _check_paths()
        else:
            if self.direc_in == 'backward':
                self.finalpath[self.direc_in].reverse()
            return self.finalpath[self.direc_in]

    ###########################
    # Classify Algorithm Type #
    ###########################
    def name_algorithm(self):
        if self.HF == self.HB > self.U or self.direc_in == 'forward':
            logging.info('Monodirectional forward labeling algorithm')
        elif self.L < self.HF == self.HB < self.U:
            logging.info(
                'Bidirectional labeling algorithm with static halfway point')
        elif self.HF == self.HB < self.L or self.direc_in == 'backward':
            logging.info('Monodirectional backward labeling algorithm')
        elif self.U == self.HF > self.HB == self.L:
            logging.info('Bidirectional labeling algorithm with dynamic' +
                         ' halfway point.')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cspy",
    version="0.0.1",
    description="A collection of algorithms for the (Resource) Constrained" +
    "Shortest Path Problem",
    license="MIT",
    author="David Torres",
    author_email="d.torressanchez@lancs.ac.uk",
    keywords=["shortest path", "resource constrained shortest path",
              "bidirectional algorithm", "tabu search"],
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/torressa/cspy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
